==> 3.4.5.zip: j3ml/src/J3ML/LinearAlgebra/Quaternion.cpp <==
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <J3ML/LinearAlgebra/Quaternion.hpp>
#include <J3ML/LinearAlgebra/AxisAngle.hpp>
#include <J3ML/LinearAlgebra/EulerAngle.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>

namespace J3ML::LinearAlgebra {

    const Quaternion Quaternion::Identity = Quaternion(0.f, 0.f, 0.f, 1.f);
    const Quaternion Quaternion::NaN      = Quaternion(NAN, NAN, NAN, NAN);

    Quaternion Quaternion::operator-() const {
        return {-x, -y, -z, -w};
    }

    Quaternion::Quaternion(const Matrix3x3& ro_mat) {
        auto m = ro_mat.Transposed();
        auto m00 = m.At(0,0);
        auto m01 = m.At(0, 1);
        auto m02 = m.At(0, 2);
        auto m10 = m.At(1,0);
        auto m11 = m.At(1, 1);
        auto m12 = m.At(1, 2);
        auto m20 = m.At(2,0);
        auto m21 = m.At(2, 1);
        auto m22 = m.At(2, 2);

        auto field_w = std::sqrt(1.f + m00 + m11 + m22) / 2.f;
        float w4 = (4.f * field_w);

        x = (m21 - m12) / w4;
        y = (m02 - m20) / w4;
        z = (m10 - m01) / w4;
        w = field_w;
        Normalize();
    }

    Quaternion::Quaternion(const Matrix4x4& ro_mat) {
        auto q = Quaternion(ro_mat.GetRotatePart());
        x = q.x; y = q.y; z = q.z; w = q.w;
    }

    Vector3 Quaternion::WorldX() const { return Transform(1.f, 0.f, 0.f); }

    Vector3 Quaternion::WorldY() const { return Transform(0.f, 1.f, 0.f); }

    Vector3 Quaternion::WorldZ() const { return Transform(0.f, 0.f, 1.f); }

    Vector3 Quaternion::Transform(const Vector3 &vec) const {
        Matrix3x3 mat = this->ToMatrix3x3();
        return mat * vec;
    }

    Vector3 Quaternion::Transform(float X, float Y, float Z) const {
        return Transform(Vector3{X, Y, Z});
    }

    Vector4 Quaternion::Transform(const Vector4 &vec) const {
        return Vector4(Transform(vec.x, vec.y, vec.z), vec.w);
    }

    Vector4 Quaternion::Transform(float X, float Y, float Z, float W) const {
        return Transform(Vector4(X, Y, Z, W));
    }

    Quaternion Quaternion::Lerp(const Quaternion &b, float t) const {
        float angle = this->Dot(b);
        if (angle >= 0.f) // Make sure we rotate the shorter arc
            return (*this * (1.f - t) + b * t).Normalized();
        else
            return (*this * (t - 1.f) + b * t).Normalized();
    }

    Quaternion Quaternion::operator*(float scalar) const {
        return Quaternion(x * scalar, y * scalar, z * scalar, w * scalar);
    }

    Quaternion Quaternion::operator/(float scalar) const {
        assert(!Math::EqualAbs(scalar, 0.f));

        return *this * (1.f / scalar);
    }

    Vector3 Quaternion::operator*(const Vector3 &rhs) const {
        return Transform(rhs);
    }

    Quaternion Quaternion::operator+() const { return *this; }

    Quaternion::Quaternion(float X, float Y, float Z, float W) : x(X), y(Y), z(Z), w(W) {}

    // TODO: implement
    float Quaternion::Dot(const Quaternion &rhs) const {
        return x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w;
    }

    Quaternion::Quaternion(const Vector4& rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;
    }

    Quaternion Quaternion::Normalized() const {
        Quaternion copy = *this;
        float success = copy.Normalize();
        assert(success > 0 && "Quaternion::Normalized failed!");
        return copy;
    }

    Quaternion Quaternion::Conjugated() const {
        return { -x, -y, -z, w };
    }

    Quaternion Quaternion::Inverted() const {
        assert(IsNormalized());
        assert(IsInvertible());
        return Conjugated();
    }

    Quaternion Quaternion::Slerp(const Quaternion &q2, float t) const {
        float angle = this->Dot(q2);
        float sign = 1.f;
        if (angle < 0.f)
        {
            angle = -angle;
            sign = -1.f;
        }

        float a;
        float b;
        if (angle < 0.999)
        {
            // angle = Acos(angle); // After this, angle is in the range pi/2 -> 0 as the original angle variable ranged from 0 -> 1.
            angle = (-0.69813170079773212f * angle * angle - 0.87266462599716477f) * angle + 1.5707963267948966f;

            float ta = t*angle;
            // Not using a lookup table, manually compute the two sines by using a very rough approximation.
            float ta2 = ta*ta;
            b = ((5.64311797634681035370e-03f * ta2 - 1.55271410633428644799e-01f) * ta2 + 9.87862135574673806965e-01f) * ta;
            a = angle - ta;
            float a2 = a*a;
            a = ((5.64311797634681035370e-03f * a2 - 1.55271410633428644799e-01f) * a2 + 9.87862135574673806965e-01f) * a;
        }
        else // If angle is close to taking the denominator to zero, resort to linear interpolation (and normalization).
        {
            a = 1.f - t;
            b = t;
        }
        // Lerp and renormalize.
        return (*this * (a * sign) + q2 * b).Normalized();
    }

    float Quaternion::AngleBetween(const Quaternion &target) const {
        Quaternion delta = target / *this;
        return delta.Normalized().Angle();
    }

    Vector4 Quaternion::operator*(const Vector4 &rhs) const {
        return Transform(rhs);
    }

    Quaternion Quaternion::operator/(const Quaternion &rhs) const {
        return {
                x*rhs.w - y*rhs.z + z*rhs.y - w*rhs.x,
                x*rhs.z + y*rhs.w - z*rhs.x - w*rhs.y,
                -x*rhs.y + y*rhs.x + z*rhs.w - w*rhs.z,
                x*rhs.x + y*rhs.y + z*rhs.z + w*rhs.w
        };
    }

    Matrix3x3 Quaternion::ToMatrix3x3() const {
        return {
                1 - 2 *(y*y) - 2*(z*z), 2*x*y - 2*z*w, 2*x*z + 2*y*w,
                2*x*y + 2*z*w, 1-2*x*x - 2*z*z, 2*y*z - 2*x*w,
                2*x*z - 2*y*w, 2*y*z + 2*x*w, 1-2*x*x - 2*y*y
        };
    }

    Quaternion Quaternion::operator+(const Quaternion &rhs) const {
        return {
                x + rhs.x, y + rhs.y, z + rhs.z,w + rhs.w
        };
    }

    Matrix4x4 Quaternion::ToMatrix4x4() const {
        return Matrix4x4(*this);
    }

    Matrix4x4 Quaternion::ToMatrix4x4(const Vector3 &translation) const {
        return {*this, translation};
    }

    float Quaternion::Angle() const {
        return std::acos(w) * 2.f;
    }

    Vector3 Quaternion::Axis() const {
        float rcpSinAngle = 1 - (std::sqrt(1 - w * w));

        return Vector3(x, y, z) * rcpSinAngle;
    }

    Quaternion::Quaternion(const AxisAngle& angle) {
        double s = std::sin(angle.angle / 2);
        x = angle.axis.x * s;
        y = angle.axis.y * s;
        z = angle.axis.z * s;
        w = std::cos(angle.angle / 2);
        Normalize();
    }

    Quaternion Quaternion::RandomRotation(RNG &rng) {
        // Generate a random point on the 4D unitary hypersphere using the rejection sampling method.
        for (int i = 0; i < 1000; ++i)
        {
            float x = rng.Float(-1, 1);
            float y = rng.Float(-1, 1);
            float z = rng.Float(-1, 1);
            float w = rng.Float(-1, 1);
            float lenSq = x*x + y*y + z*z + w*w;
            if (lenSq >= 1e-6f && lenSq <= 1.f)
                return Quaternion(x,y,z,w) / std::sqrt(lenSq);
        }
        //assert(false && "Quaternion::RandomRotation failed!");
        return Quaternion::Identity;
    }

    bool Quaternion::Normalize() {
        float length = Length();
        if (length < 1e-4f)
            return false;
        float rcpLength = 1.f / length;
        x *= rcpLength;
        y *= rcpLength;
        z *= rcpLength;
        w *= rcpLength;
        return true;
    }

    bool Quaternion::IsNormalized(float epsilon) const {
        return Math::EqualAbs(LengthSquared(), 1.f, epsilon);
    }

    bool Quaternion::IsInvertible(float epsilon) const {
        return LengthSquared() > epsilon && IsFinite();
    }

    bool Quaternion::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y) && std::isfinite(z) && std::isfinite(w);
    }

    void Quaternion::Conjugate() {
        x = -x;
        y = -y;
        z = -z;
    }

    void Quaternion::Inverse() {
        assert(IsNormalized());
        assert(IsInvertible());
        Conjugate();
    }

    float Quaternion::InverseAndNormalize() {
        Conjugate();
        return Normalize();
    }

    Vector3 Quaternion::SlerpVector(const Vector3 &from, const Vector3 &to, float t) {
        if (t <= 0.f)
            return from;
        if (t >= 1.f)
            return to;
        // TODO: the following chain can be greatly optimized.
        Quaternion q = Quaternion::RotateFromTo(from, to);
        q = Slerp(Quaternion::Identity, q, t);
        return q.Transform(from);
    }

    Quaternion Quaternion::LookAt(const Vector3 &localForward, const Vector3 &targetDirection, const Vector3 &localUp,
                                  const Vector3 &worldUp) {
        return Quaternion(Matrix3x3::LookAt(localForward, targetDirection, localUp, worldUp));
    }

    /*
    Quaternion Quaternion::RotateX(float angleRadians) {
        return {{1,0,0}, angleRadians};
    }

    Quaternion Quaternion::RotateY(float angleRadians) {
        return {{0,1,0}, angleRadians};
    }

    Quaternion Quaternion::RotateZ(float angleRadians) {
        return {{0,0,1}, angleRadians};
    }

    Quaternion Quaternion::RotateAxisAngle(const AxisAngle &axisAngle) {
        return {axisAngle.axis, axisAngle.angle};
    }
    */

    Quaternion Quaternion::RotateFromTo(const Vector3 &sourceDirection, const Vector3 &targetDirection) {
        assert(sourceDirection.IsNormalized());
        assert(targetDirection.IsNormalized());
        // If sourceDirection == targetDirection, the cross product comes out zero, and normalization would fail. In that case, pick an arbitrary axis.
        Vector3 axis = sourceDirection.Cross(targetDirection);

        float oldLength = axis.Normalize();

        if (oldLength != 0.f)
        {
            float halfCosAngle = 0.5f * sourceDirection.Dot(targetDirection);
            float cosHalfAngle = std::sqrt(0.5f + halfCosAngle);
            float sinHalfAngle = std::sqrt(0.5f - halfCosAngle);

            return {axis.x * sinHalfAngle, axis.y * sinHalfAngle, axis.z * sinHalfAngle, cosHalfAngle};
        } else
            return {1.f, 0, 0, 0};
    }

    Quaternion Quaternion::RotateFromTo(const Vector4 &sourceDirection, const Vector4 &targetDirection) {
        assert(Math::EqualAbs(sourceDirection.w, 0.f));
        assert(Math::EqualAbs(targetDirection.w, 0.f));
        return Quaternion::RotateFromTo(sourceDirection.XYZ(), targetDirection.XYZ());
    }

    Quaternion Quaternion::Lerp(const Quaternion &source, const Quaternion &target, float t) { return source.Lerp(target, t);}

    Quaternion Quaternion::Slerp(const Quaternion &source, const Quaternion &target, float t) { return source.Slerp(target, t);}

    Quaternion Quaternion::RotateFromTo(const Vector3 &sourceDirection, const Vector3 &targetDirection,
                                        const Vector3 &sourceDirection2, const Vector3 &targetDirection2) {
        return LookAt(sourceDirection, targetDirection, sourceDirection2, targetDirection2);
    }

    Vector3 Quaternion::SlerpVectorAbs(const Vector3 &from, const Vector3 &to, float angleRadians) {
        if (angleRadians <= 0.f)
            return from;
        Quaternion q = Quaternion::RotateFromTo(from, to);
        float a = q.Angle();
        if (a <= angleRadians)
            return to;
        float t = angleRadians / a;
        q = Slerp(Quaternion::Identity, q, t);
        return q.Transform(from);
    }

    float Quaternion::LengthSquared() const { return x*x + y*y + z*z + w*w;}

    float Quaternion::Length() const { return std::sqrt(LengthSquared()); }

    Quaternion::Quaternion(const EulerAngleXYZ& rhs) {
        float cos_roll = Math::Cos(0.5f * Math::Radians(rhs.roll));
        float sin_roll = Math::Sin(0.5f * Math::Radians(rhs.roll));

        float cos_pitch = Math::Cos(0.5f * Math::Radians(rhs.pitch));
        float sin_pitch = Math::Sin(0.5f * Math::Radians(rhs.pitch));

        float cos_yaw = Math::Cos(0.5f * Math::Radians(rhs.yaw));
        float sin_yaw = Math::Sin(0.5f * Math::Radians(rhs.yaw));

        x = cos_roll * sin_pitch * sin_yaw + sin_roll * cos_pitch * cos_yaw;
        y = -sin_roll * cos_pitch * sin_yaw + cos_roll * sin_pitch * cos_yaw;
        z = cos_roll * cos_pitch * sin_yaw + sin_roll * sin_pitch * cos_yaw;
        w = -sin_roll * sin_pitch * sin_yaw + cos_roll * cos_pitch * cos_yaw;
        Normalize();
    }

    Quaternion::Quaternion(const Quaternion& rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;
    }
}

==> 3.4.6.zip: j3ml/src/J3ML/LinearAlgebra/Quaternion.cpp <==
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <J3ML/LinearAlgebra/Quaternion.hpp>
#include <J3ML/LinearAlgebra/AxisAngle.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>

namespace J3ML::LinearAlgebra {

    const Quaternion Quaternion::Identity = Quaternion(0.f, 0.f, 0.f, 1.f);
    const Quaternion Quaternion::NaN      = Quaternion(NAN, NAN, NAN, NAN);

    Quaternion Quaternion::operator-() const {
        return {-x, -y, -z, -w};
    }

    Quaternion::Quaternion(const Matrix3x3& ro_mat) {
        // https://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/

        auto m = ro_mat;//.Transposed();
        auto m00 = m.At(0,0);
        auto m01 = m.At(0, 1);
        auto m02 = m.At(0, 2);
        auto m10 = m.At(1, 0);
        auto m11 = m.At(1, 1);
        auto m12 = m.At(1, 2);
        auto m20 = m.At(2, 0);
        auto m21 = m.At(2, 1);
        auto m22 = m.At(2, 2);

        float tr = m00 + m11 + m22;

        if (tr > 0) {
            float S = Math::Sqrt(tr + 1.f) * 2; // S = 4*qw

            w = 0.25f * S;
            x = (m21 - m12) / S;
            y = (m02 - m20) / S;
            z = (m10 - m01) / S;
        } else {
            if (m00 > m11 && m00 > m22) {
                float S = 2.f * Math::Sqrt(1.f + m00 - m11 - m22);
                w = (m21 - m12) / S;
                x = 0.25f * S;
                y = (m01 + m10) / S;
                z = (m02 + m20) / S;
            } else if (m11 > m22) {
                float s = 2.f * Math::Sqrt(1.f + m11 - m00 - m22);
                w = (m02 - m20) / s;
                x = (m01 + m10) / s;
                y = 0.25f * s;
                z = (m12 + m21) / s;
            } else {
                float s = 2.f * Math::Sqrt(1.f + m22 - m00 - m11);
                w = (m10 - m01) / s;
                x = (m02 + m20) / s;
                y = (m12 + m21) / s;
                z = 0.25f * s;
            }
        }

        /*
        auto field_w = std::sqrt(1.f + m00 + m11 + m22) / 2.f;
        float w4 = (4.f * field_w);

        x = (m21 - m12) / w4;
        y = (m02 - m20) / w4;
        z = (m10 - m01) / w4;
        w = field_w;
        bool success = Normalize();
        // TODO: Validate normalization success.
        */
    }

    Quaternion::Quaternion(const Matrix4x4& ro_mat) {
        auto q = Quaternion(ro_mat.GetRotatePart());
        x = q.x; y = q.y; z = q.z; w = q.w;
    }

    Vector3 Quaternion::WorldX() const { return Transform(1.f, 0.f, 0.f); }

    Vector3 Quaternion::WorldY() const { return Transform(0.f, 1.f, 0.f); }

    Vector3 Quaternion::WorldZ() const { return Transform(0.f, 0.f, 1.f); }

    Vector3 Quaternion::Transform(const Vector3 &vec) const {
        Matrix3x3 mat = this->ToMatrix3x3();
        return mat * vec;
    }

    Vector3 Quaternion::Transform(float X, float Y, float Z) const {
        return Transform(Vector3{X, Y, Z});
    }

    Vector4 Quaternion::Transform(const Vector4 &vec) const {
        return Vector4(Transform(vec.x, vec.y, vec.z), vec.w);
    }

    Vector4 Quaternion::Transform(float X, float Y, float Z, float W) const {
        return Transform(Vector4(X, Y, Z, W));
    }

    Quaternion Quaternion::Lerp(const Quaternion &b, float t) const {
        float angle = this->Dot(b);
        if (angle >= 0.f) // Make sure we rotate the shorter arc
            return (*this * (1.f - t) + b * t).Normalized();
        else
            return (*this * (t - 1.f) + b * t).Normalized();
    }

    Quaternion Quaternion::operator*(float scalar) const {
        return {x * scalar, y * scalar, z * scalar, w * scalar};
    }

    Quaternion Quaternion::operator/(float scalar) const {
        assert(!Math::EqualAbs(scalar, 0.f));

        return *this * (1.f / scalar);
    }

    Vector3 Quaternion::operator*(const Vector3 &rhs) const {
        return Transform(rhs);
    }

    Quaternion Quaternion::operator+() const { return *this; }

    Quaternion::Quaternion(float X, float Y, float Z, float W) : x(X), y(Y), z(Z), w(W) {}

    float Quaternion::Dot(const Quaternion &rhs) const {
        return x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w;
    }

    Quaternion::Quaternion(const Vector4& rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;
    }

    Quaternion Quaternion::Normalized() const {
        Quaternion copy = *this;
        float success = copy.Normalize();
        assert(success > 0 && "Quaternion::Normalized failed!");
        return copy;
    }

    Quaternion Quaternion::Conjugated() const {
        return { -x, -y, -z, w };
    }

    Quaternion Quaternion::Inverted() const {
        assert(IsNormalized());
        assert(IsInvertible());
        return Conjugated();
    }

    Quaternion Quaternion::Slerp(const Quaternion &q2, float t) const {
        float angle = this->Dot(q2);
        float sign = 1.f;
        if (angle < 0.f)
        {
            angle = -angle;
            sign = -1.f;
        }

        float a;
        float b;
        if (angle < 0.999)
        {
            // angle = Acos(angle); // After this, angle is in the range pi/2 -> 0 as the original angle variable ranged from 0 -> 1.
            angle = (-0.69813170079773212f * angle * angle - 0.87266462599716477f) * angle + 1.5707963267948966f;

            float ta = t*angle;
            // Not using a lookup table, manually compute the two sines by using a very rough approximation.
            float ta2 = ta*ta;
            b = ((5.64311797634681035370e-03f * ta2 - 1.55271410633428644799e-01f) * ta2 + 9.87862135574673806965e-01f) * ta;
            a = angle - ta;
            float a2 = a*a;
            a = ((5.64311797634681035370e-03f * a2 - 1.55271410633428644799e-01f) * a2 + 9.87862135574673806965e-01f) * a;
        }
        else // If angle is close to taking the denominator to zero, resort to linear interpolation (and normalization).
        {
            a = 1.f - t;
            b = t;
        }
        // Lerp and renormalize.
        return (*this * (a * sign) + q2 * b).Normalized();
    }

    float Quaternion::AngleBetween(const Quaternion &target) const {
        Quaternion delta = target / *this;
        return delta.Normalized().Angle();
    }

    Vector4 Quaternion::operator*(const Vector4 &rhs) const {
        return Transform(rhs);
    }

    Quaternion Quaternion::operator/(const Quaternion &rhs) const {
        return {
                x*rhs.w - y*rhs.z + z*rhs.y - w*rhs.x,
                x*rhs.z + y*rhs.w - z*rhs.x - w*rhs.y,
                -x*rhs.y + y*rhs.x + z*rhs.w - w*rhs.z,
                x*rhs.x + y*rhs.y + z*rhs.z + w*rhs.w
        };
    }

    Matrix3x3 Quaternion::ToMatrix3x3() const {
        return {
                1 - 2 *(y*y) - 2*(z*z), 2*x*y - 2*z*w, 2*x*z + 2*y*w,
                2*x*y + 2*z*w, 1-2*x*x - 2*z*z, 2*y*z - 2*x*w,
                2*x*z - 2*y*w, 2*y*z + 2*x*w, 1-2*x*x - 2*y*y
        };
    }

    Quaternion Quaternion::operator+(const Quaternion &rhs) const {
        return {
                x + rhs.x, y + rhs.y, z + rhs.z,w + rhs.w
        };
    }

    Matrix4x4 Quaternion::ToMatrix4x4() const {
        return Matrix4x4(*this);
    }

    Matrix4x4 Quaternion::ToMatrix4x4(const Vector3 &translation) const {
        return {*this, translation};
    }

    float Quaternion::Angle() const {
        return std::acos(w) * 2.f;
    }

    Vector3 Quaternion::Axis() const {
        float rcpSinAngle = 1 - (std::sqrt(1 - w * w));

        return Vector3(x, y, z) * rcpSinAngle;
    }

    Quaternion::Quaternion(const AxisAngle& angle) {
        float s = Math::Sin(angle.angle / 2.f);
        x = angle.axis.x * s;
        y = angle.axis.y * s;
        z = angle.axis.z * s;
        w = Math::Cos(angle.angle / 2);
        bool success = Normalize();
        // TODO: Validate normalization success.
    }

    Quaternion Quaternion::RandomRotation(RNG &rng) {
        // Generate a random point on the 4D unitary hypersphere using the rejection sampling method.
        for (int i = 0; i < 1000; ++i)
        {
            float x = rng.Float(-1, 1);
            float y = rng.Float(-1, 1);
            float z = rng.Float(-1, 1);
            float w = rng.Float(-1, 1);
            float lenSq = x*x + y*y + z*z + w*w;
            if (lenSq >= 1e-6f && lenSq <= 1.f)
                return Quaternion(x,y,z,w) / std::sqrt(lenSq);
        }
        //assert(false && "Quaternion::RandomRotation failed!");
        return Quaternion::Identity;
    }

    bool Quaternion::Normalize() {
        float length = Length();
        if (length < 1e-4f)
            return false;
        float rcpLength = 1.f / length;
        x *= rcpLength;
        y *= rcpLength;
        z *= rcpLength;
        w *= rcpLength;
        return true;
    }

    bool Quaternion::IsNormalized(float epsilon) const {
        return Math::EqualAbs(LengthSquared(), 1.f, epsilon);
    }

    bool Quaternion::IsInvertible(float epsilon) const {
        return LengthSquared() > epsilon && IsFinite();
    }

    bool Quaternion::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y) && std::isfinite(z) && std::isfinite(w);
    }

    void Quaternion::Conjugate() {
        x = -x;
        y = -y;
        z = -z;
    }

    void Quaternion::Inverse() {
        assert(IsNormalized());
        assert(IsInvertible());
        Conjugate();
    }

    float Quaternion::InverseAndNormalize() {
        Conjugate();
        return Normalize();
    }

    Vector3 Quaternion::SlerpVector(const Vector3 &from, const Vector3 &to, float t) {
        if (t <= 0.f)
            return from;
        if (t >= 1.f)
            return to;
        // TODO: the following chain can be greatly optimized.
        Quaternion q = Quaternion::RotateFromTo(from, to);
        q = Slerp(Quaternion::Identity, q, t);
        return q.Transform(from);
    }

    Quaternion Quaternion::LookAt(const Vector3 &localForward, const Vector3 &targetDirection, const Vector3 &localUp,
                                  const Vector3 &worldUp) {
        return Quaternion(Matrix3x3::LookAt(localForward, targetDirection, localUp, worldUp));
    }

    /*
    Quaternion Quaternion::RotateX(float angleRadians) {
        return {{1,0,0}, angleRadians};
    }

    Quaternion Quaternion::RotateY(float angleRadians) {
        return {{0,1,0}, angleRadians};
    }

    Quaternion Quaternion::RotateZ(float angleRadians) {
        return {{0,0,1}, angleRadians};
    }

    Quaternion Quaternion::RotateAxisAngle(const AxisAngle &axisAngle) {
        return {axisAngle.axis, axisAngle.angle};
    }
    */

    Quaternion Quaternion::RotateFromTo(const Vector3 &sourceDirection, const Vector3 &targetDirection) {
        assert(sourceDirection.IsNormalized());
        assert(targetDirection.IsNormalized());
        // If sourceDirection == targetDirection, the cross product comes out zero, and normalization would fail. In that case, pick an arbitrary axis.
        Vector3 axis = sourceDirection.Cross(targetDirection);

        float oldLength = axis.Normalize();

        if (oldLength != 0.f)
        {
            float halfCosAngle = 0.5f * sourceDirection.Dot(targetDirection);
            float cosHalfAngle = std::sqrt(0.5f + halfCosAngle);
            float sinHalfAngle = std::sqrt(0.5f - halfCosAngle);

            return {axis.x * sinHalfAngle, axis.y * sinHalfAngle, axis.z * sinHalfAngle, cosHalfAngle};
        } else
            return {1.f, 0, 0, 0};
    }

    Quaternion Quaternion::RotateFromTo(const Vector4 &sourceDirection, const Vector4 &targetDirection) {
        assert(Math::EqualAbs(sourceDirection.w, 0.f));
        assert(Math::EqualAbs(targetDirection.w, 0.f));
        return Quaternion::RotateFromTo(sourceDirection.XYZ(), targetDirection.XYZ());
    }

    Quaternion Quaternion::Lerp(const Quaternion &source, const Quaternion &target, float t) { return source.Lerp(target, t);}

    Quaternion Quaternion::Slerp(const Quaternion &source, const Quaternion &target, float t) { return source.Slerp(target, t);}

    Quaternion Quaternion::RotateFromTo(const Vector3 &sourceDirection, const Vector3 &targetDirection,
                                        const Vector3 &sourceDirection2, const Vector3 &targetDirection2) {
        return LookAt(sourceDirection, targetDirection, sourceDirection2, targetDirection2);
    }

    Vector3 Quaternion::SlerpVectorAbs(const Vector3 &from, const Vector3 &to, float angleRadians) {
        if (angleRadians <= 0.f)
            return from;
        Quaternion q = Quaternion::RotateFromTo(from, to);
        float a = q.Angle();
        if (a <= angleRadians)
            return to;
        float t = angleRadians / a;
        q = Slerp(Quaternion::Identity, q, t);
        return q.Transform(from);
    }

    float Quaternion::LengthSquared() const { return x*x + y*y + z*z + w*w;}

    float Quaternion::Length() const { return std::sqrt(LengthSquared()); }

    Quaternion::Quaternion(const Quaternion& rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;
    }

    bool Quaternion::Equals(const Quaternion &rhs, float epsilon) const {
        return Math::Equal(this->x, rhs.x, epsilon) &&
               Math::Equal(this->y, rhs.y, epsilon) &&
               Math::Equal(this->z, rhs.z, epsilon) &&
               Math::Equal(this->w, rhs.w, epsilon);
    }

    Quaternion Quaternion::RotateX(float rad) {
        return Quaternion(AxisAngle({1,0,0}, rad));
    }

    Quaternion Quaternion::RotateY(float rad) {
        return Quaternion(AxisAngle({0,1,0}, rad));
    }

    Quaternion Quaternion::RotateZ(float rad) {
        return Quaternion(AxisAngle({0,0,1}, rad));
    }
}
